/*
 * Copyright (C) Cross The Road Electronics.  All rights reserved.
 * License information can be found in CTRE_LICENSE.txt
 * For support and suggestions contact support@ctr-electronics.com or file
 * an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
 */
#pragma once

#include "ctre/phoenix6/configs/Configs.hpp"
#include "ctre/phoenix6/controls/ControlRequest.hpp"
#include "ctre/phoenix6/networking/interfaces/Control_Interface.h"
#include <sstream>
#include <units/angular_velocity.h>
#include <units/angle.h>
#include <units/frequency.h>
#include <units/time.h>


namespace ctre {
namespace phoenix6 {
namespace controls {

/**
 * Request PID to target velocity with a differential position setpoint.
 *
 * This control mode will set the motor's velocity setpoint to the velocity specified by the user. It will
 * also set the motor's differential position setpoint to the specified position.
 */
class DifferentialVelocityDutyCycle : public ControlRequest
{
    bool ApplyConfigsOnRequest{false};

    ctre::phoenix::StatusCode SendRequest(const char *network, uint32_t deviceHash, bool cancelOtherRequests, std::shared_ptr<ControlRequest> &req)
    {
        std::stringstream ss;
        auto& ref = requestReference.GetNameValues();
        ss << TargetVelocity.to<double>(); ref["TargetVelocity"] = ss.str(); ss.str(std::string{});
        ss << DifferentialPosition.to<double>(); ref["DifferentialPosition"] = ss.str(); ss.str(std::string{});
        ss << EnableFOC; ref["EnableFOC"] = ss.str(); ss.str(std::string{});
        ss << TargetSlot; ref["TargetSlot"] = ss.str(); ss.str(std::string{});
        ss << DifferentialSlot; ref["DifferentialSlot"] = ss.str(); ss.str(std::string{});
        ss << OverrideBrakeDurNeutral; ref["OverrideBrakeDurNeutral"] = ss.str(); ss.str(std::string{});

        if (req.get() != this)
        {
            auto const reqCast = dynamic_cast<DifferentialVelocityDutyCycle *>(req.get());
            if (reqCast != nullptr)
            {
                *reqCast = *this;
            }
            else
            {
                req = std::make_shared<DifferentialVelocityDutyCycle>(*this);
            }
        }

        
        std::string strs{ss.str()};
        c_ctre_phoenix6_requestConfigApply(network, deviceHash, ConfigTimeout.to<double>(), strs.c_str(), strs.length(), ApplyConfigsOnRequest);
        ApplyConfigsOnRequest = false;
        return c_ctre_phoenix6_RequestControlDifferentialVelocityDutyCycle(network, deviceHash, UpdateFreqHz.to<double>(), cancelOtherRequests, TargetVelocity.to<double>(), DifferentialPosition.to<double>(), EnableFOC, TargetSlot, DifferentialSlot, OverrideBrakeDurNeutral);
    }

public:
    /**
     * Average velocity to drive toward in rotations per second.
     */
    units::angular_velocity::turns_per_second_t TargetVelocity;
    /**
     * Differential position to drive toward in rotations.
     */
    units::angle::turn_t DifferentialPosition;
    /**
     * Set to true to use FOC commutation (requires Phoenix Pro), which increases
     * peak power by ~15%. Set to false to use trapezoidal commutation.  FOC
     * improves motor performance by leveraging torque (current) control.  However,
     * this may be inconvenient for applications that require specifying duty cycle
     * or voltage.  CTR-Electronics has developed a hybrid method that combines the
     * performances gains of FOC while still allowing applications to provide duty
     * cycle or voltage demand.  This not to be confused with simple sinusoidal
     * control or phase voltage control which lacks the performance gains.
     */
    bool EnableFOC;
    /**
     * Select which gains are applied to the primary controller by selecting the
     * slot.  Use the configuration api to set the gain values for the selected slot
     * before enabling this feature. Slot must be within [0,2].
     */
    int TargetSlot;
    /**
     * Select which gains are applied to the differential controller by selecting
     * the slot.  Use the configuration api to set the gain values for the selected
     * slot before enabling this feature. Slot must be within [0,2].
     */
    int DifferentialSlot;
    /**
     * Set to true to static-brake the rotor when output is zero (or within
     * deadband).  Set to false to use the NeutralMode configuration setting
     * (default). This flag exists to provide the fundamental behavior of this
     * control when output is zero, which is to provide 0V to the motor.
     */
    bool OverrideBrakeDurNeutral;

    

    /**
     * \brief The timeout when sending configs associated with this control
     */
    units::time::second_t ConfigTimeout{0.1_s};

    /**
     * \brief The period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     */
    units::frequency::hertz_t UpdateFreqHz{100_Hz}; // Default to 100_Hz

    /**
     *\brief  Request PID to target velocity with a differential position setpoint.
     *
     *\details  This control mode will set the motor's velocity setpoint to the
     *          velocity specified by the user. It will also set the motor's
     *          differential position setpoint to the specified position.
     *
     *\param TargetVelocity     Average velocity to drive toward in rotations per
     *                          second.
     *\param DifferentialPosition     Differential position to drive toward in
     *                                rotations.
     *\param EnableFOC     Set to true to use FOC commutation (requires Phoenix
     *                     Pro), which increases peak power by ~15%. Set to false to
     *                     use trapezoidal commutation.  FOC improves motor
     *                     performance by leveraging torque (current) control. 
     *                     However, this may be inconvenient for applications that
     *                     require specifying duty cycle or voltage. 
     *                     CTR-Electronics has developed a hybrid method that
     *                     combines the performances gains of FOC while still
     *                     allowing applications to provide duty cycle or voltage
     *                     demand.  This not to be confused with simple sinusoidal
     *                     control or phase voltage control which lacks the
     *                     performance gains.
     *\param TargetSlot     Select which gains are applied to the primary controller
     *                      by selecting the slot.  Use the configuration api to set
     *                      the gain values for the selected slot before enabling
     *                      this feature. Slot must be within [0,2].
     *\param DifferentialSlot     Select which gains are applied to the differential
     *                            controller by selecting the slot.  Use the
     *                            configuration api to set the gain values for the
     *                            selected slot before enabling this feature. Slot
     *                            must be within [0,2].
     *\param OverrideBrakeDurNeutral     Set to true to static-brake the rotor when
     *                                   output is zero (or within deadband).  Set
     *                                   to false to use the NeutralMode
     *                                   configuration setting (default). This flag
     *                                   exists to provide the fundamental behavior
     *                                   of this control when output is zero, which
     *                                   is to provide 0V to the motor.
     */
    DifferentialVelocityDutyCycle(units::angular_velocity::turns_per_second_t TargetVelocity, units::angle::turn_t DifferentialPosition, bool EnableFOC, int TargetSlot, int DifferentialSlot, bool OverrideBrakeDurNeutral) : ControlRequest{"DifferentialVelocityDutyCycle"},
        TargetVelocity{std::move(TargetVelocity)},
        DifferentialPosition{std::move(DifferentialPosition)},
        EnableFOC{std::move(EnableFOC)},
        TargetSlot{std::move(TargetSlot)},
        DifferentialSlot{std::move(DifferentialSlot)},
        OverrideBrakeDurNeutral{std::move(OverrideBrakeDurNeutral)}
    {}

        /**
     *\brief  Request PID to target velocity with a differential position setpoint.
     *
     *\details  This control mode will set the motor's velocity setpoint to the
     *          velocity specified by the user. It will also set the motor's
     *          differential position setpoint to the specified position.
     *
     *\param TargetVelocity     Average velocity to drive toward in rotations per
     *                          second.
     *\param DifferentialPosition     Differential position to drive toward in
     *                                rotations.
     */
    DifferentialVelocityDutyCycle(units::angular_velocity::turns_per_second_t TargetVelocity, units::angle::turn_t DifferentialPosition) : DifferentialVelocityDutyCycle{TargetVelocity, DifferentialPosition, true, 0, 1, false}
    {}
    
    /**
     * \brief Modifies this Control Request's TargetVelocity parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newTargetVelocity Parameter to modify
     * \returns Itself
     */
    DifferentialVelocityDutyCycle& WithTargetVelocity(units::angular_velocity::turns_per_second_t newTargetVelocity)
    {
        TargetVelocity = std::move(newTargetVelocity);
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's DifferentialPosition parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newDifferentialPosition Parameter to modify
     * \returns Itself
     */
    DifferentialVelocityDutyCycle& WithDifferentialPosition(units::angle::turn_t newDifferentialPosition)
    {
        DifferentialPosition = std::move(newDifferentialPosition);
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's EnableFOC parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newEnableFOC Parameter to modify
     * \returns Itself
     */
    DifferentialVelocityDutyCycle& WithEnableFOC(bool newEnableFOC)
    {
        EnableFOC = std::move(newEnableFOC);
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's TargetSlot parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newTargetSlot Parameter to modify
     * \returns Itself
     */
    DifferentialVelocityDutyCycle& WithTargetSlot(int newTargetSlot)
    {
        TargetSlot = std::move(newTargetSlot);
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's DifferentialSlot parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newDifferentialSlot Parameter to modify
     * \returns Itself
     */
    DifferentialVelocityDutyCycle& WithDifferentialSlot(int newDifferentialSlot)
    {
        DifferentialSlot = std::move(newDifferentialSlot);
        return *this;
    }
    
    /**
     * \brief Modifies this Control Request's OverrideBrakeDurNeutral parameter and returns itself for
     *        method-chaining and easier to use request API.
     * \param newOverrideBrakeDurNeutral Parameter to modify
     * \returns Itself
     */
    DifferentialVelocityDutyCycle& WithOverrideBrakeDurNeutral(bool newOverrideBrakeDurNeutral)
    {
        OverrideBrakeDurNeutral = std::move(newOverrideBrakeDurNeutral);
        return *this;
    }
    /**
     * \brief Sets the period at which this control will update at.
     * This is designated in Hertz, with a minimum of 20 Hz
     * (every 50 ms) and a maximum of 1000 Hz (every 1 ms).
     *
     * If this field is set to 0 Hz, the control request will
     * be sent immediately as a one-shot frame. This may be useful
     * for advanced applications that require outputs to be
     * synchronized with data acquisition. In this case, we
     * recommend not exceeding 50 ms between control calls.
     *
     * \param newUpdateFreqHz Parameter to modify
     * \returns Itself
     */
    DifferentialVelocityDutyCycle &WithUpdateFreqHz(units::frequency::hertz_t newUpdateFreqHz)
    {
        UpdateFreqHz = newUpdateFreqHz;
        return *this;
    }
    /**
     * Returns a string representation of the object.
     *
     * \returns a string representation of the object.
     */
    std::string ToString() const
    {
        std::stringstream ss;
        ss << "class: DifferentialVelocityDutyCycle" << std::endl;
        ss << "TargetVelocity: " << TargetVelocity.to<double>() << std::endl;
        ss << "DifferentialPosition: " << DifferentialPosition.to<double>() << std::endl;
        ss << "EnableFOC: " << EnableFOC << std::endl;
        ss << "TargetSlot: " << TargetSlot << std::endl;
        ss << "DifferentialSlot: " << DifferentialSlot << std::endl;
        ss << "OverrideBrakeDurNeutral: " << OverrideBrakeDurNeutral << std::endl;
        return ss.str();
    }

    /**
     * \brief Forces configs to be applied the next time this is used in a setControl.
     *        This is not necessary in the majority of cases, because Phoenix will make sure configs are
     *        properly set when they are not already set
     */
    void ForceApplyConfigs() { ApplyConfigsOnRequest = true; }
};

}
}
}

